/* Copyright 2016-2024 Michele Santullo
 * This file is part of "duckhandy".
 *
 * "duckhandy" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * "duckhandy" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with "duckhandy".  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef idFC25566D624140559C54B39FFFE52F04
#define idFC25566D624140559C54B39FFFE52F04

#include <type_traits>
#include <stdexcept>
#if !defined(INT_CONV_WITHOUT_HELPERS)
#	include <string_view>
#	include <string>
#	include <sstream>
#endif
#include <tuple>

namespace dhandy {
	template <typename T, std::size_t S>
	class ReversedSizedArray {
		static_assert(S > 0, "This container requires size to be at least 1");
		static_assert(std::is_trivial<T>::value, "Only use this container with trivial types");
	public:
		constexpr static const std::size_t capacity = S;

		using iterator = T*;
		constexpr ReversedSizedArray() = default;
		~ReversedSizedArray() = default;

		constexpr std::size_t size() const { return S - (m_curr + 1); }
		constexpr bool empty() const { return m_curr + 1 == S; }
		constexpr const T operator[] (std::size_t idx) const { if (idx >= size()) throw std::out_of_range("Out of bound array access"); return m_data[idx + m_curr + 1]; }
		constexpr T& operator[] (std::size_t idx) { if (idx >= size()) throw std::out_of_range("Out of bound array access"); return m_data[idx + m_curr + 1]; }
		constexpr void push_front (const T& itm) { if (size() == S) throw std::length_error("ReversedSizedArray is full"); m_data[m_curr--] = itm; }
		constexpr const T* data() const { return m_data + m_curr + 1; }
		constexpr const T* base_ptr() const { return m_data; }
		constexpr iterator begin() { return m_data + m_curr + 1; }
		constexpr iterator end() { return m_data + S; }
		constexpr const T& back() const { return *(m_data + S - 1); }

		template <typename V>
		constexpr V to() const { return V(data(), size() - (not empty() and not back() ? 1 : 0)); }
		constexpr auto to_tuple() const;

#if !defined(INT_CONV_WITHOUT_HELPERS)
		constexpr std::basic_string_view<T> to_string_view() const { return to<std::basic_string_view<T>>(); }
		bool operator== (const std::basic_string_view<T>& other) const { return to_string_view() == other; }
		bool operator!= (const std::basic_string_view<T>& other) const { return not operator==(other); }
		bool operator== (const std::basic_string<T>& other) const { return to_string_view() == other; }
		bool operator!= (const std::basic_string<T>& other) const { return not operator==(other); }
		bool operator== (const T* other) const { return to_string_view() == std::basic_string_view<T>(other); }
		bool operator!= (const T* other) const { return not operator==(other); }

		std::string to_string() const { return to<std::basic_string<T>>(); }
#endif

	private:
		T m_data[S];
		std::size_t m_curr {S - 1};
	};

	namespace implem {
		template <typename T, std::size_t S, std::size_t... Indices>
		constexpr auto elements_or_empty_to_tuple (const T (&elems)[S], std::size_t from_idx, std::index_sequence<Indices...>) {
			return std::make_tuple((from_idx + Indices < S ? elems[from_idx + Indices] : T{})...);
		}
	} //namespace implem

#if !defined(INT_CONV_WITHOUT_HELPERS)
	template <typename T, std::size_t S>
	inline
	std::basic_ostream<T>& operator<< (std::basic_ostream<T>& stream, const ReversedSizedArray<T, S>& arr) {
		stream << arr.to_string_view();
		return stream;
	}
#endif

	template <typename T, std::size_t S>
	constexpr auto ReversedSizedArray<T, S>::to_tuple() const {
		return implem::elements_or_empty_to_tuple(m_data, m_curr + 1, std::make_index_sequence<S>{});
	}
} //namespace dhandy

#endif
