/* Copyright 2016-2024 Michele Santullo
 * This file is part of "duckhandy".
 *
 * "duckhandy" is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * "duckhandy" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with "duckhandy".  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef id890F1A22A90047BE9A9DCC8F7F057F91
#define id890F1A22A90047BE9A9DCC8F7F057F91

#include "implem/int_conv.hpp"
#include <type_traits>
#include <string>
#include <string_view>
#include <array>

#define DHANDY_INTCONV_USE_STDSTRING

namespace dhandy {
	namespace implem {
		template <typename T, typename F, typename Tr, bool FromInt=std::is_integral_v<F>>
		struct IntConv;

		template <typename Tr>
		using IntConvString = std::basic_string<typename Tr::char_type>;

		template <typename Tr>
		using IntConvStringView = std::basic_string_view<typename Tr::char_type>;

		template <typename F, typename Tr>
		struct IntConv<IntConvString<Tr>, F, Tr, true> {
			static IntConvString<Tr> conv (const F& in) {
				auto retval = dhandy::int_to_ary<F, 10, Tr>(in);
				return IntConvString<Tr>{retval.begin(), retval.end() - 1};
			}
		};
#if defined(DHANDY_INTCONV_USE_STDSTRING)
		template <typename F, typename C>
		struct IntConv<IntConvString<ASCIITranslator<C>>, F, ASCIITranslator<C>, true> {
			static IntConvString<ASCIITranslator<C>> conv (const F& in) {
				return std::to_string(in);
			}
		};
#endif
		template <typename F, typename Tr>
		struct IntConv<IntConvStringView<Tr>, F, Tr, true> {
			consteval static IntConvStringView<Tr> conv (const F& in) {
				return dhandy::int_to_ary<F, 10, Tr>(in).to_string_view();
			}
		};
		template <typename F, typename Tr, typename C, std::size_t N>
		struct IntConv<std::array<C, N>, F, Tr, true> {
		private:
			template <typename... Args, std::size_t... Indices>
			constexpr static std::array<C, N> to_array (const std::tuple<Args...>& vals, std::index_sequence<Indices...>) {
				return std::array<C, N>{std::get<Indices>(vals)...};
			}
		public:
			constexpr static std::array<C, N> conv (const F& in) {
				return to_array(
					dhandy::int_to_ary<F, 10, Tr>(in).to_tuple(),
					std::make_index_sequence<
						std::min(decltype(dhandy::int_to_ary<F, 10, Tr>(in))::capacity, N)
					>{}
				);
			}
		};
		template <typename T, typename Tr>
		struct IntConv<T, IntConvString<Tr>, Tr, false> {
			constexpr static T conv (const IntConvString<Tr>& in) {
				return dhandy::ary_to_int<T, char, 10, Tr>(
					in.data(),
					in.data() + in.size()
				);
			}
		};
		template <typename T, typename Tr>
		struct IntConv<T, IntConvStringView<Tr>, Tr, false> {
			constexpr static T conv (const IntConvStringView<Tr>& in) {
				return dhandy::ary_to_int<T, char, 10, Tr>(
					in.data(),
					in.data() + in.size()
				);
			}
		};
		template <typename T, typename F, F In, typename Tr, bool FromInt>
		struct IntConv<T, std::integral_constant<F, In>, Tr, FromInt> {
			constexpr static T conv (const std::integral_constant<F, In>&) {
				return dhandy::buildtime_int_to_ary<F, In, 10, Tr>().to_string_view();
			}
		};
	} //namespace implem

	template <typename To, typename From, bool AllCaps=false>
	constexpr inline To int_conv (const From& from, std::integral_constant<bool, AllCaps> = std::integral_constant<bool, AllCaps>{}) {
		if constexpr (AllCaps)
			return implem::IntConv<To, From, ASCIITranslatorUpcase<char>>::conv(from);
		else
			return implem::IntConv<To, From, ASCIITranslator<char>>::conv(from);
	}

	template <typename From>
	inline auto int_conv_raw (const From& from) {
		return int_to_ary<From, 10, ASCIITranslator<char>>(from);
	}
} //namespace dhandy

#endif
